
# ngrams - given a study carrel and an integer, output... ngrams

# require
from rdr import *

# ngrams
@click.command( options_metavar='<options>' )
@click.option('-q', '--query', type=click.STRING, help="limit results to include the given word")
@click.option('-c', '--count', is_flag=True, help='count and tabulate the result')
@click.option('-l', '--location', default='local', type=click.Choice( [ 'local', 'remote' ] ), help='where is the library')
@click.option('-s', '--size', default=1, help='output in a more human-readable form')
@click.argument( 'carrel', metavar='<carrel>' )
def ngrams( carrel, size, query, count, location ) :

	"""Output ngrams of the given <size> from <carrel>"""

	# require
	from re       import search
	from requests import get
	import nltk

	# branch according to location; local
	if location == 'local' :
	
		# sanity check
		checkForCarrel( carrel )
		
		# read local data
		localLibrary = configuration( 'localLibrary' )
		stopwords    = open( str( localLibrary/carrel/ETC/STOPWORDS ) ).read().split()
		text         = open( str( localLibrary/carrel/ETC/CORPUS ) ).read()
	
	# remote
	elif location == 'remote' :
	
		# read remote data; needs error checking
		text      = get( '/'.join ( [ REMOTELIBRARY, CARRELS, carrel, ETC, CORPUS ] ) ).text
		stopwords = get( '/'.join ( [ REMOTELIBRARY, CARRELS, carrel, ETC, STOPWORDS ] ) ).text.split()
		
	# error
	else :
	
		# click ought to prevent this, but just in case
		click.echo( "Error: Unknown value for location: { location }. Call Eric.", err=True )
		exit()
			
	# read, tokenize, and normalize the text
	tokens = nltk.word_tokenize( text, preserve_line=True )
	tokens = [ token.lower() for token in tokens if token.isalpha() ]
	
	# create the set of ngrams
	ngrams = list( nltk.ngrams( tokens, size ) )
	
	# filter, conditionally
	if query :
	
		# initialize and process each ngram
		filtered = []
		for ngram in ngrams :
			
			# check and update
			if search( query, ' '.join( ngram ) ) : filtered.append( ngram )
		
		# done
		ngrams = filtered

	# remove stopwords from unigrams or bigrams
	if size < 3 :
	
		# initialize
		results = []
		
		# process each ngram
		for ngram in ngrams :

			# re-initialize
			found = False
			
			# process each token in the ngram
			for token in ngram :

				# check for stopword
				if token in stopwords : found = True
	
			# conditionally update the results
			if not found : results.append( ngram )

		# done; make it read pretty
		ngrams = results
		
	# output; human
	if count :
	
		# initialize a dictionary and process each ngram
		items = {}
		for ngram in ngrams :

			# update the dictionary
			if ngram in items : items[ ngram ] += 1
			else              : items[ ngram ]  = 1

		# sort the dictionary; return the ngrams
		ngrams = sorted( items.items(), key=lambda x:x[ 1 ], reverse=True )
		
		# process each ngram
		for ngram in ngrams :
			
			# create a record and output
			record = str( ngram[ 1 ] ) + '\t' + '\t'.join( list( ngram[ 0 ] ) )
			click.echo( record )
		
	# power user
	else :
	
		# output raw data, and hope sort, uniq, grep, and less are used
		for ngram in ngrams : click.echo( "\t".join( list( ngram ) ) )





# grammars - output different types of features found in a text

# require
from rdr import *


# handle model not found error
def modelNotFound() :
	
	# notify
	click.echo( "Error: Langauge model not found.", err=True )
	click.echo()
	click.echo( f"This functions requires a spaCy langauge model ({ MODEL}) to be installed. This only has to be done once, and after the model has been installed you can run the command again.", err=True )
	click.echo()
	click.echo( 'Do you want to install the model now? [yn] ', err=True, nl=False )
	
	# get input
	c = click.getchar()
	click.echo()
	
	# branch accordingly; yes
	if c == 'y' :

		# require and do the work
		from os import system
		system( 'python -m spacy download ' + MODEL )
	
	# no
	elif c == 'n' : click.echo( "Okay, but installing the model is necessary for this function to work. You'll be asked again next time.", err=True )

	# error
	else : click.echo( '???' )
	
	# done
	exit()


# given a carrel, return a spacy doc
def carrel2doc( carrel ) :

	# configure
	PICKLE = 'reader.spacy'

	# require
	from os        import path, stat
	from spacy     import load
	import textacy
	
	# initialize
	localLibrary = configuration( 'localLibrary' )
	pickle       = localLibrary/carrel/ETC/PICKLE

	# check to see if we've previously been here
	if path.exists( pickle ) :
		
		# read the pickle file
		try            : doc = next( textacy.io.spacy.read_spacy_docs( pickle, lang=MODEL ) )
		except OSError : modelNotFound()
			
	# otherwise
	else :
	
		# warn
		click.echo( '''Modeling study carrel data for future use. This may take many
minutes, but it will only have to be done once. In the meantime,
ask yourself, "Self, what is justice?"''', err=True )

		# initialize 
		file           = localLibrary/carrel/ETC/CORPUS
		text           = open( str( file ) ).read()
		size           = ( stat( file ).st_size ) + 1
		
		# initialize some more
		try            : nlp  = load( MODEL )
		except OSError : modelNotFound()
		
		# do the work
		nlp.max_length = size
		doc            = nlp( text )

		# save it for future use
		textacy.io.spacy.write_spacy_docs( doc, filepath=pickle )

	# done
	return doc


# filter
def applyFilter ( query, features ) :

	# require
	from re import search
	
	# initialize and process each feature
	filtered = []
	for feature in features :
				
		# branch according to type; span
		if str( type( feature ) ) == "<class 'spacy.tokens.span.Span'>" : string = feature.text

		# hope for a tuple with three elements; these need to be functions
		else :
		
			if str( type( feature ) ) == "<class 'textacy.extract.triples.SVOTriple'>" :		

				# parse and stringify
				subject = [ token.text_with_ws for token in feature.subject ]
				verb    = [ token.text_with_ws for token in feature.verb ]
				object  = [ token.text_with_ws for token in feature.object ]
				string  = ( ' '.join( [ ''.join( subject ), ''.join( verb ), ''.join( object ) ] ) )

			elif str( type( feature ) ) == "<class 'textacy.extract.triples.DQTriple'>" :		

				# parse and stringify
				speaker = [ token.text_with_ws for token in feature.speaker ]
				cue     = [ token.text_with_ws for token in feature.cue ]
				content = feature.content.text_with_ws
				string  = ( ' '.join( [ ''.join( speaker ), ''.join( cue ), content ] ) )

			elif str( type( feature ) ) == "<class 'textacy.extract.triples.SSSTriple'>" :		

				# parse and stringify
				entity   = [ token.text_with_ws for token in feature.entity ]
				cue      = [ token.text_with_ws for token in feature.cue ]
				fragment = [ token.text_with_ws for token in feature.fragment ]
				string  = ( ' '.join( [ ''.join( entity ), ''.join( cue ), ''.join( fragment ) ] ) )

			# error
			else :
			
				# get the type of feature, output, and quit
				feature = str( type( feature ) )
				click.echo( f"Error: Unknown value for type of feature { feature }. Call Eric.", err=True )
				exit()

		# do the work
		if ( search( query, string ) ) : filtered.append( feature )

	# done
	return ( filtered )


# output
def outputFeatures( features ) :

	# process each feature
	for feature in features :
			
		# branch accordingly; nouns
		if str( type( feature ) ) == "<class 'spacy.tokens.span.Span'>" : click.echo( feature.text )
		
		# subject-verb-object; svo
		elif str( type( feature ) ) == "<class 'textacy.extract.triples.SVOTriple'>" :
		
			# parse
			subject = [ token.text_with_ws for token in feature.subject ]
			verb    = [ token.text_with_ws for token in feature.verb ]
			object  = [ token.text_with_ws for token in feature.object ]
						
			# output; a bit obtuse
			click.echo( '\t'.join( [ ''.join( subject ), ''.join( verb ), ''.join( object ) ] ) )

		# direct quotes
		elif str( type( feature ) ) == "<class 'textacy.extract.triples.DQTriple'>" :
		
			# parse
			speaker = [ token.text_with_ws for token in feature.speaker ]
			cue     = [ token.text_with_ws for token in feature.cue ]
			content = feature.content.text_with_ws
			
			# output; still a bit obtuse
			click.echo( '\t'.join( [ ''.join( speaker ), ''.join( cue ), content ] ) )
			
		# direct quotes
		elif str( type( feature ) ) == "<class 'textacy.extract.triples.SSSTriple'>" :
		
			# parse
			entity   = [ token.text_with_ws for token in feature.entity ]
			cue      = [ token.text_with_ws for token in feature.cue ]
			fragment = [ token.text_with_ws for token in feature.fragment ]
			
			# output; a bit obtuse some more
			click.echo( '\t'.join( [ ''.join( entity ), ''.join( cue ), ''.join( fragment ) ] ) )
		
		# error
		else :
			
			# get the type of feature, output, and quit
			feature = str( type( feature ) )
			click.echo( f"Error: Unknown value for type of feature { feature }. Call Eric.", err=True )
			exit()


# grammars
@click.command( options_metavar='<options>' )
@click.option('-q', '--query', type=click.STRING, help="filter results using the given regular expression")
@click.option('-n', '--noun',  type=click.STRING, help="only applicable to sss; a noun or noun phrase")
@click.option('-l', '--lemma', default='be', help="only applicable to sss; the lemma of a verb, such as 'be' (default), 'have', or 'say'")
@click.option('-g', '--grammar', default='svo', type=click.Choice( [ 'svo', 'sss', 'nouns', 'quotes' ], case_sensitive=True ), help="only applicable to sss; the lemma of a verb, such as 'be' (default), 'have', or 'say'")
@click.argument( 'carrel', metavar='<carrel>' )
def grammars( carrel, grammar, query, noun, lemma ) :

	"""Extract grammatical sentence fragments from <carrel> where fragments are one of:
	
	\b
	  * nouns - all the nouns and noun chunks
	  * quotes - things people say
	  * svo - fragments in the form of subject-verb-object
	  * sss - a more advanced version of svo; fragments beginning
	    with an entity, are connected with a verb, and are followed
	    by a phrase
	
	Examples:
	
	\b
	  * rdr grammars homer svo
	  * rdr grammars -q horse homer svo
	"""

	# require
	from textacy import extract
	from os      import system
		
	# sanity check
	checkForCarrel( carrel )

	# initialize
	doc = carrel2doc( carrel )

	# get the features
	if   grammar == 'svo'    : features = list( extract.subject_verb_object_triples( doc ) )
	elif grammar == 'quotes' : features = list( extract.direct_quotations( doc ) )
	elif grammar == 'nouns'  : features = list( extract.noun_chunks( doc ) )
	elif grammar == 'sss'    :

		# sanity check
		if not noun :
		
			click.echo( "Error: When specifying sss, the -n option is required. See 'rdr grammars --help'.", err=True )
			exit()
			
		# do the work
		features = list( extract.semistructured_statements( doc, entity=noun, cue=lemma ) )

	# filter, conditionally
	if query : features = applyFilter( query, features )
	
	# output
	outputFeatures( features )
	





# require
from rdr import *

# config
@click.command()
@click.argument( 'item', type=click.Choice( [ 'local', 'mallet' ], case_sensitive=False ) )
def set( item ) :

	"""Set the location of various items"""
	
	# require
	from configparser import ConfigParser
	from pathlib      import Path
	
	# initialize
	configurations       = ConfigParser()
	applicationDirectory = Path( click.get_app_dir( APPLICATIONDIRECTORY ) )
	configurationFile    = applicationDirectory/CONFIGURATIONFILE

	# create configuration file, conditionally
	if not configurationFile.exists() :
		
		# initialize
		configurations[ "RDR" ] = { "localLibrary"  : Path.home()/READERLIBRARY, "malletHome" : '' }

		# create directory and save the file
		applicationDirectory.mkdir( parents=False, exist_ok=True )
		with open( str( configurationFile ), 'w' ) as handle : configurations.write( handle )
		
	# re-initialize
	localLibrary  = configuration( 'localLibrary' )
	malletHome    = configuration( 'malletHome' )

	# branch accordingly, local
	if item == 'local' :
	
		# get the desired library location
		click.echo( 'Where do you want to save your study carrels? Press enter to accept the default.' )
		localLibrary = input( 'Directory [%s]: ' % localLibrary ) or localLibrary
		localLibrary = Path( localLibrary )

		# try to create the directory and save the configuration
		try : localLibrary.mkdir( exist_ok=True )
		except FileNotFoundError : click.echo( "Error: File not found. Are you sure you entered a valid path?", err=True )		

	# mallet
	elif item == 'mallet' :
	
		# get the desired library location
		click.echo( 'What is the full path to your MALLET distribution?' )
		malletHome = input( 'Directory [%s]: ' % malletHome ) or malletHome

	# update the configuration file
	configurations[ "RDR" ] = { "localLibrary"  : localLibrary, "malletHome" : malletHome }
	with open( str( configurationFile ), 'w' ) as handle : configurations.write( handle )






# require
from rdr import *
 
def hangman() :

	# https://inventwithpython.com/invent4thed/chapter8.html

	import random
	HANGMAN_PICS = ['''
  +---+
      |
      |
      |
     ===''', '''
  +---+
  O   |
      |
      |
     ===''', '''
  +---+
  O   |
  |   |
      |
     ===''', '''
  +---+
  O   |
 /|   |
      |
     ===''', '''
  +---+
  O   |
 /|\  |
      |
     ===''', '''
  +---+
  O   |
 /|\  |
 /    |
     ===''', '''
  +---+
  O   |
 /|\  |
 / \  |
     ===''']

	words = 'ant baboon badger bat bear beaver camel cat clam cobra cougar coyote crow deer dog donkey duck eagle ferret fox frog goat goose hawk lion lizard llama mole monkey moose mouse mule newt otter owl panda parrot pigeon python rabbit ram rat raven rhino salmon seal shark sheep skunk sloth snake spider stork swan tiger toad trout turkey turtle weasel whale wolf wombat zebra'.split()

	def getRandomWord(wordList):
	 # This function returns a random string from the passed list of strings.
	 wordIndex = random.randint(0, len(wordList) - 1)
	 return wordList[wordIndex]

	def displayBoard(missedLetters, correctLetters, secretWord):
	 print(chr(27) + "[2J")
	 print(HANGMAN_PICS[len(missedLetters)])
	 print()

	 print('Missed letters:', end=' ')
	 for letter in missedLetters:
		 print(letter, end=' ')
	 print()

	 blanks = '_' * len(secretWord)

	 for i in range(len(secretWord)): # Replace blanks with correctly guessed letters.
		 if secretWord[i] in correctLetters:
			 blanks = blanks[:i] + secretWord[i] + blanks[i+1:]

	 for letter in blanks: # Show the secret word with spaces in between each letter.
		 print(letter, end=' ')
	 print()

	def getGuess(alreadyGuessed):
	 # Returns the letter the player entered. This function makes sure the player entered a single letter and not something else.
	 while True:
		 print('Guess a letter.')
		 guess = input()
		 guess = guess.lower()
		 if len(guess) != 1:
			 print('Please enter a single letter.')
		 elif guess in alreadyGuessed:
			 print('You have already guessed that letter. Choose again.')
		 elif guess not in 'abcdefghijklmnopqrstuvwxyz':
			 print('Please enter a LETTER.')
		 else:
			 return guess

	def playAgain():
	 # This function returns True if the player wants to play again; otherwise, it returns False.
	 print('Do you want to play again? (yes or no)')
	 return input().lower().startswith('y')


	print('H A N G M A N')
	missedLetters = ''
	correctLetters = ''
	secretWord = getRandomWord(words)
	gameIsDone = False

	while True:
	 displayBoard(missedLetters, correctLetters, secretWord)

	 # Let the player enter a letter.
	 guess = getGuess(missedLetters + correctLetters)

	 if guess in secretWord:
		 correctLetters = correctLetters + guess

		 # Check if the player has won.
		 foundAllLetters = True
		 for i in range(len(secretWord)):
			 if secretWord[i] not in correctLetters:
				 foundAllLetters = False
				 break
		 if foundAllLetters:
			 print('Yes! The secret word is "' + secretWord +
				   '"! You have won!')
			 gameIsDone = True
	 else:
		 missedLetters = missedLetters + guess

		 # Check if player has guessed too many times and lost.
		 if len(missedLetters) == len(HANGMAN_PICS) - 1:
			 displayBoard(missedLetters, correctLetters, secretWord)
			 print('You have run out of guesses!\nAfter ' +
				   str(len(missedLetters)) + ' missed guesses and ' + str(len(correctLetters)) + ' correct guesses, the word was "' + secretWord + '"')
			 gameIsDone = True

	 # Ask the player if they want to play again (but only if the game is done).
	 if gameIsDone:
		 if playAgain():
			 missedLetters = ''
			 correctLetters = ''
			 gameIsDone = False
			 secretWord = getRandomWord(words)
		 else:
			 break


# config
@click.command()
@click.argument( 'game' )
def play( game ) :

	"""Play a game"""

	if   game == 'hangman' : hangman()
	else : 
		click.echo( f"Error: Unknown value for GAME: { game }" )
		os.system( 'rdr play --help' )		






# rdr.py - a command-line interface for using Distant Reader study carrels

# Eric Lease Morgan <emorgan@nd.edu>
# (c) University of Notre Dame; distributed under a GNU Public License

# July 30, 2021 - in Three Oaks with Pat; first real working version

# require
from rdr import *
import rdr.commands.browse      as browse
import rdr.commands.cluster     as cluster
import rdr.commands.concordance as concordance
import rdr.commands.sql   as sql
import rdr.commands.download    as download
import rdr.commands.get         as get
import rdr.commands.grammars    as grammars
import rdr.commands.catalog     as catalog
import rdr.commands.ngrams      as ngrams
import rdr.commands.play        as play
import rdr.commands.read        as read
import rdr.commands.edit         as edit
import rdr.commands.set         as set
import rdr.commands.tm          as tm

# initialize 
@click.group()
def rdr() : pass

# update the list of commands
rdr.add_command( browse.browse )
rdr.add_command( cluster.cluster )
rdr.add_command( concordance.concordance )
rdr.add_command( sql.sql )
rdr.add_command( download.download )
rdr.add_command( get.get )
rdr.add_command( edit.edit )
rdr.add_command( grammars.grammars )
rdr.add_command( catalog.catalog )
rdr.add_command( ngrams.ngrams )
rdr.add_command( play.play )
rdr.add_command( read.read )
rdr.add_command( set.set )
rdr.add_command( tm.tm )

# do the work
if __name__ == '__main__' : rdr()




# concordance - given a carrel and a word, output rudimentary keyword-in-context

# require
from rdr import *

@click.command( options_metavar='[<options>]' )
@click.option('-w', '--width', default=80, help='number of characters in each line of output')
@click.option('-l', '--lines', default=999, help='number of lines of text to output')
@click.option('-q', '--query', default='love', help='output in a more human-readable form')
@click.argument( 'carrel', metavar='<carrel>' )
def concordance( carrel, query, width, lines ) :

	"""Output matching lines from <carrel> where <query> is a word or phrase
	
	Developed in the 12th Century, concordancing is the oldest of text mining tools and the poor man's search engine. Sometimes it is called a keyword-in-context (KWIC) index. Given <query>, this is a quick and easy way to see how it is used in the same breath as other words. Use 'rdr ngrams' to identify words or phrases of interest.
	
	Examples:
	
	\b
	  * rdr concordance homer hector
	  * rdr concordance homer 'hector was'
	
	See also: rdr ngrams"""
	
	# require
	from nltk import Text, word_tokenize

	# sanity checks
	checkForCarrel( carrel )
	checkForPunkt()
	
	# initialize, read, and normalize; ought to save the result for future use
	localLibrary = configuration( 'localLibrary' )
	corpus       = localLibrary/carrel/ETC/CORPUS
	text         = Text( word_tokenize( open( corpus ).read( ) ) )
					
	# split query into a list, conditionally
	if ' ' in query : query = query.split( ' ' )
		
	# do the work and output
	lines = text.concordance_list( query, width=width, lines=lines )
	for line in lines : click.echo( line.line )
	



# browse - given a carrel, use lynx to navigate its file system

# require
from rdr import *

@click.command( options_metavar='[<options>]' )
@click.option('-l', '--location', default='local', type=click.Choice( [ 'local', 'remote' ] ), help='where is the library')
@click.argument( 'carrel', metavar='<carrel>' )
def browse( carrel, location ) :

	"""Use a Web browser called Lynx to peruse <carrel>
	
	Study carrels are sets of HTML files, other plain text files, and a whole lot of tab-delimited files organized into a file system -- a data set. Lynx is very useful tool for inspecting the contents of these file systems from the terminal, especially if you are not accessing the carrel through a Web server.
	
	Example: rdr browse homer
	
	See also: rdr read"""

	# configure
	LYNX = 'lynx'
	
	# require
	from os import system
	import webbrowser
	
	# local carrel
	if location == 'local' :
	
		# sanity check
		checkForCarrel( carrel )

		# initialize, create a URL, and do the work
		localLibrary = str( configuration( 'localLibrary' ) )
		url          = 'file://' + localLibrary + '/' + carrel
		system( LYNX + ' ' + url )

	# remote carrel
	elif location == 'remote' :
	
		url = '/'.join ( [ REMOTELIBRARY, CARRELS, carrel, MANIFEST ] )
		webbrowser.open( url )




# read - open the root of a study carrel in a Web browser

# require
from rdr import *

# read
@click.command( options_metavar='<options>' )
@click.option('-l', '--location', default='local', type=click.Choice( [ 'local', 'remote' ] ), help='output in a more human-readable form')
@click.argument( 'carrel', metavar='<carrel>' )
def read( carrel, location ) :

	"""Open <carrel> in your default Web browser where the location is one of:
	
	\b
	  * local - from your local library
	  * remote - from the remote library
	
	Example: rdr read homer
	"""
		
	# require
	from webbrowser import open
	
	if location == 'local' :
	
		# sanity check
		checkForCarrel( carrel )

		localLibrary  = configuration( 'localLibrary' )
		url = 'file://' + str( localLibrary/carrel/INDEX )
		open( url )
		
	elif location == 'remote' :
	
		url = REMOTELIBRARY + '/' + CARRELS + '/' + carrel
		open( url )




# require
from rdr import *

# config
@click.command()
@click.option('-s', '--setting', default='local', type=click.Choice( [ 'local', 'mallet' ] ), help='output the given setting')
def get( setting ) :

	"""Echo the full path to your local library of study carrels
	
	This is useful if you want to change directories to your local collection of carrels, and then interact with them directly.
	
	Examples:
	
	\b
	  * rdr home"""
	
	# branch accordingly
	if   setting == 'local' : click.echo( str( configuration( 'localLibrary' ) ) )
	elif setting == 'mallet' : click.echo( str( configuration( 'malletHome' ) ) )
	
	




# download - given the name of a remote carrel, cache it locally

# require
from rdr import *

# harvest
@click.command( options_metavar='[<options>]' )
@click.argument( 'carrel', metavar='<carrel>' )
def download( carrel ) :

	"""Download <carrel> from the public library of study carrels
	
	A collection of about 3,000 pre-created study carrels is available at http://library.distantreader.org. Given <carrel>, this function will download the remote carrel and cache it in your local library. Use 'rdr list remote' to learn what carrels are available.
	
	Example: rdr download author-homer-gutenberg
	
	See also: rdr list"""
			
	# configure
	ZIPFILE = 'study-carrel.zip'
	CARRELS = 'carrels'

	# require
	from requests import get
	from tempfile import TemporaryFile
	from zipfile  import ZipFile

	# initialize
	localLibrary  = configuration( 'localLibrary' )

	# get the remote zip file; needs error checking
	click.echo( "Getting study carrel... ", err=True )
	response = get( REMOTELIBRARY + '/' + CARRELS + '/' + carrel + '/' + ZIPFILE )
	
	# initialize a temporary file and write to it
	click.echo( "Saving study carrel... ", err=True )
	handle = TemporaryFile()
	handle.write( response.content )
	
	# unzip the temporary file and close it, which also deletes it
	click.echo( "Unziping study carrel... " )
	with ZipFile( handle, 'r' ) as zip : zip.extractall( str( localLibrary ) )
	handle.close()

	# done
	click.echo( "Done." )




# catalog - given a location, output a list of study carrels

# require
from rdr import *

@click.command( options_metavar='<options>' )
@click.option('-h', '--human', is_flag=True, help='output in a more human-readable form')
@click.option('-l', '--location', default='local', type=click.Choice( [ 'local', 'remote' ] ), help='output in a more human-readable form')
def catalog( human, location ) :

	"""List the items in a library where <location> is either 'local' or 'remote'
	
	Use this function to learn what items are saved in your local library of study carrels or what items are available in the library of about 3,000 pre-created study carrels at http://library.distantreader.org/. The local option returns a simple list. The remote option returns a tab-delimited list along with some metadata describing each item. In either case, you might want to pipe the results to utilities such as sort, grep, cut, or less in order to make the most of the result.
	
	Examples:
	
	\b
	  * rdr list local
	  * rdr list remote | less -S
	  * rdr list -h remote
	  * rdr list remote | grep love | less -S -x32
	  * rdr list remote | grep love | cut -f1 | less
	  
	See also: rdr download"""
	
	# configure
	TSV    = 'catalog/catalog.tsv'
	RECORD = "      item: ##ITEM##\n      name: ##NAME##\n      date: ##DATE##\n  keywords: ##KEYWORDS##\n     items: ##ITEMS##\n     words: ##WORDS##\n     score: ##SCORE##\n     bytes: ##BYTES##\n\n"
	HEADER = "\nThe catalog includes ##COUNT## items, and each is listed below:\n\n"

	# require
	from requests import get
		
	# branch accordingly; local
	if location == 'local' :
		
		# initialize
		localLibrary = configuration( 'localLibrary' )
		
		# read, sort, and output
		carrels = [ carrel.name for carrel in localLibrary.iterdir() if carrel.is_dir() ]
		carrels.sort()
		for carrel in carrels : click.echo( carrel )
	
	# remote
	elif location == 'remote' :
	
		# create person-amenable output
		if human :
		
			# create a rudimentary catalog
			catalog = ''
			count   = 0
			
			records = get( REMOTELIBRARY + '/' + TSV ).text 
			for item, record in enumerate( records.split( '\n' ) ) :
			
				# delimit and sanity check
				fields = record.split( '\t' )
				if len( fields ) != 7 : break
			
				# parse
				name     = fields[ 0 ]
				date     = fields[ 1 ]
				keywords = fields[ 2 ]
				items    = fields[ 3 ]
				words    = fields[ 4 ]
				score    = fields[ 5 ]
				bytes    = fields[ 6 ]
			
				# increment
				count += 1
				item  += 1
				
				# update
				record   = RECORD.replace( '##ITEM##', str( item ) )
				record   = record.replace( '##NAME##', name )
				record   = record.replace( '##DATE##', date )
				record   = record.replace( '##KEYWORDS##', keywords )
				record   = record.replace( '##ITEMS##', items )
				record   = record.replace( '##WORDS##', words )
				record   = record.replace( '##SCORE##', score )
				record   = record.replace( '##BYTES##', bytes )
				catalog += record
				
			# add the header and output
			header  = HEADER.replace( '##COUNT##', str( count ) )
			catalog = header + catalog
			click.echo_via_pager( catalog )
				
		# get the raw data and hope the results get piped to utilities like sort, grep, cut, less, etc.
		else : click.echo( get( REMOTELIBRARY + '/' + TSV ).text  )





# tm - do topic modeling

# require
from rdr import *

def checkForMallet( mallet ) :
	
	# require
	from configparser import ConfigParser
	from pathlib      import Path
	from requests     import get
	from tempfile     import TemporaryFile
	from zipfile      import ZipFile
	
	# install mallet, conditionally
	if not Path( mallet ).exists() :
	
		click.echo( "MALLET not found. Downoading MALLET... ", err=True )
		response = get( MALLETZIP )

		# initialize a temporary file and write to it
		click.echo( "Saving zip file... ", err=True )
		handle = TemporaryFile()
		handle.write( response.content )

		# unzip the temporary file and close it, which also deletes it
		click.echo( "Unziping zip file... " )
		with ZipFile( handle, 'r' ) as zip : zip.extractall( Path.home() )
		
		# initialize
		click.echo( "Updating configurations... " )
		configurations          = ConfigParser()
		applicationDirectory    = Path( click.get_app_dir( APPLICATIONDIRECTORY ) )
		configurationFile       = applicationDirectory/CONFIGURATIONFILE
		localLibrary            = configuration( 'localLibrary' )
		malletHome              = Path.home()/'mallet'
		configurations[ "RDR" ] = { "localLibrary"  : localLibrary, "malletHome" : malletHome }
		with open( str( configurationFile ), 'w' ) as handle : configurations.write( handle )

		# make mallet executable
		click.echo( "Making MALLET executable... " )
		(malletHome/MALLETBIN).chmod( 0x755 )
		

		# done
		click.echo('''Done. MALLET has been downloaded to your home directory and configured
for future use. You can move MALLET to another location but once you do
so you will need to run 'rdr set -s mallet'.''', err=True )


@click.command( options_metavar='<options>' )
@click.option('-t', '--topics', default=7, help="number of topics to generate")
@click.option('-w', '--words', default=30, help="number of words used to describe topic")
@click.option('-i', '--iterations', default=2400, help="number of iterations")
@click.argument( 'carrel', metavar='<carrel>' )
def tm( carrel, topics, words, iterations ) :

	"""Apply topic modeling against <carrel>
	
	Topic modeling is the process of enumerating latent themes from a corpus, and it is yet another way to describe a corpus's aboutness. It is suggested you start out small when it comes to the values for --topics and --words. Repeat the modeling process an gradually increase the values."
	
	Example: rdr tm homer -t 3 -d 3
	"""

	# configure
	HTML       = 'topic-model.htm'
	RANDOMSEED = 42
	WORKERS    = 6

	# require
	from gensim                 import corpora
	from gensim.models.wrappers import LdaMallet, ldamallet
	from nltk                   import word_tokenize
	from os                     import environ
	from pathlib                import Path
	import pyLDAvis
	import pyLDAvis.gensim_models
	import webbrowser

	# sanity checks
	checkForCarrel( carrel )
	checkForMallet( str( configuration( 'malletHome' ) ) + '/' + MALLETBIN )
	checkForPunkt()

	# initialize
	localLibrary = configuration( 'localLibrary' )
	mallet       = str( configuration( 'malletHome' ) ) + '/' + MALLETBIN
	stopwords    = open( localLibrary/carrel/ETC/STOPWORDS ).read().split()
	files        = Path( localLibrary/carrel/TXT ).glob( '*.txt' )

	# create a list of normalized texts
	texts = []
	for file in files :

		tokens = open( file ).read()
		tokens = word_tokenize( tokens )
		tokens = [ token.lower() for token in tokens if token.isalpha() ]
		tokens = [ token for token in tokens if token not in stopwords ]
		texts.append( tokens )

	# create a dictionary and term-document matrix
	id2word = corpora.Dictionary( texts )
	corpus  = [ id2word.doc2bow( text ) for text in texts ]

	# use mallet to model the data and convert it to something gensim can understand
	model = LdaMallet( mallet, corpus=corpus, num_topics=topics, id2word=id2word, workers=WORKERS, iterations=iterations, random_seed=RANDOMSEED )
	model = ldamallet.malletmodel2ldamodel( model )

	# create, save, and open the visualization
	html          = str( localLibrary/carrel/HTM/HTML )
	visualization = pyLDAvis.gensim_models.prepare( model, corpus, id2word, sort_topics=True, R=words )
	pyLDAvis.save_html( visualization, html )
	webbrowser.open( 'file://' + html )



"""
a command-line tool for interacting with Distant Reader study carrels

rdr -- shorthand for Reader Toolbox -- is a command-line tool for
interacting with Distant Reader study carrels.

The Distant Reader (https:/distantreader.org) takes an almost arbitrary
amount of unstructured data (text) as input, does text mining and
natural language processing against it, and outputs sets of structured
data affectionatly called "study carrels".

As you may or may not know, a library study carrel is a little table or
room assigned to individual students. The students are then authorized
to collect thing from around the library, bring them to their study
carrel, organize them in any way they so desire, and persue their
research. Distant Reader study carrels are a modern-day version of the
venerable library study carrel.

This tool -- rdr -- enables the student, researcher, or scholar to
quickly and easily digest the content of carrels to address questions
from the mundane to the sublime. 

Eric Lease Morgan <emorgan@nd.edu>
(c) University of Notre Dame; distributed under a GNU Public License
"""

# configure; name of application, basename of configuration file, and default basename of local library
APPLICATIONDIRECTORY = 'rdr'
CONFIGURATIONFILE    = '.rdrrc'
READERLIBRARY        = 'reader-library'

# remote library
REMOTELIBRARY = 'http://library.distantreader.org'
CARRELS       = 'carrels'

# file system mappings
CORPUS    = 'reader.txt'
DATABASE  = 'reader.db'
ETC       = 'etc'
HTM       = 'htm'
INDEX     = 'index.htm'
MANIFEST  = 'MANIFEST.xml'
STOPWORDS = 'stopwords.txt'
TXT       = 'txt'

# spacy langauge model
MODEL = 'en_core_web_sm'

# mallet
MALLETZIP = 'http://library.distantreader.org/apps/mallet.zip'
MALLETBIN = 'bin/mallet'


# require
import click


# make sure the NLTK is sane
def checkForPunkt() :

	# require
	import nltk
		
	try : nltk.data.find( 'tokenizers/punkt' )
	except LookupError : 
		click.echo( "Installing punkt. This ought to only happen once.", err=True )
		nltk.download( 'punkt', quiet=True )


# make sure a study carrel exists
def checkForCarrel( carrel ) :
	
	# initialize and do the work
	directory = configuration( 'localLibrary' )/carrel
	if not directory.is_dir() :
		
		# error
		click.echo( f"The carrel, { carrel }, does not seem to be in your local library. Are you sure you entered its name correctly? Try 'rdr catalog' to make sure.", err=True )
		exit()

	
# read configurations
def configuration( name ) :

	# require
	from configparser import ConfigParser
	from pathlib      import Path

	# initialize
	applicationDirectory = Path( click.get_app_dir( APPLICATIONDIRECTORY ) )
	configurationFile    = applicationDirectory / CONFIGURATIONFILE
	configurations       = ConfigParser()
	
	# try to read configurations
	configurations.read( str( configurationFile ) )
	
	# try to get localLibrary
	try : localLibrary = configurations[ 'RDR' ][ 'localLibrary' ]
	except KeyError :
		click.echo( "Error: Key error. The location of your local study carrels has not been set. Please run 'rdr set local'.", err=True )
		exit()
		
	# try to get MALLET's home
	malletHome = configurations[ 'RDR' ][ 'malletHome' ] 
	
	# done
	if   name == 'localLibrary' : return( Path( localLibrary ) )
	elif name == 'malletHome'   : return( Path( malletHome ) )
	else :
		click.echo( f"Error: Unknown value for configuration name: { name }. Call Eric.", err=True )
		exit()
		
		




# cluster 

# require
from rdr import *

# config
@click.command( options_metavar='[<options>]' )
@click.argument( 'carrel', metavar='<carrel>' )
@click.option('-t', '--type', default='dendrogram', type=click.Choice( [ 'cube', 'dendrogram' ] ), help='output in a more human-readable form')
def cluster( carrel, type ) :

	"""Apply dimension reduction to <carrel> and visualize where <type> is 'cube' or 'dendrogram'
	
	This is a useful function to determine how holistic the contents of a study carrel is.
	
	Example: rdr cluster homer dendrogram
	
	See also: rdr mallet"""

	# configure
	MAXIMUM   = 0.95
	MINIMUM   = 2
	STOPWORDS = 'english'
	EXTENSION = '.txt'

	# require
	from os                              import path, system, listdir
	from scipy.cluster.hierarchy         import ward, dendrogram
	from sklearn.feature_extraction.text import TfidfVectorizer
	from sklearn.manifold                import MDS
	from sklearn.metrics.pairwise        import cosine_similarity
	import matplotlib.pyplot             as     plt

	# sanity check
	checkForCarrel( carrel )

	# initialize
	localLibrary = configuration( 'localLibrary' )
	directory    = localLibrary/carrel/TXT
	filenames    = [ path.join( directory, filename ) for filename in listdir( directory ) ]
	vectorizer   = TfidfVectorizer( input='filename', max_df=MAXIMUM, min_df=MINIMUM, stop_words=STOPWORDS )
	matrix       = vectorizer.fit_transform( filenames ).toarray()
	distance     = 1 - cosine_similarity( matrix )
	keys         = [ path.basename( filename ).replace( EXTENSION, '' ) for filename in filenames ] 

	# branch according to type; dendrogram
	if type == 'dendrogram' :
		linkage_matrix = ward( distance )
		dendrogram( linkage_matrix, orientation="right", labels=keys )
		plt.tight_layout() 

	# cube
	elif type == 'cube' :
		mds = MDS( n_components=3, dissimilarity="precomputed", random_state=1 )
		pos = mds.fit_transform( distance )
		fig = plt.figure()
		ax  = fig.add_subplot( 111, projection='3d' )
		ax.scatter( pos[ :, 0 ], pos[ :, 1 ], pos[ :, 2 ] )
		for x, y, z, s in zip( pos[ :, 0 ], pos[ :, 1 ], pos[ :, 2 ], keys ) : ax.text( x, y, z, s )

	# error
	else : 
		click.echo( f"Error: Unknown value for TYPE: { type }" )
		system( 'rdr cluster --help' )		

	# output
	plt.show()






from setuptools import setup, find_packages

setup(
    name='reader-toolbox',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[ 'click<7.2.0,>=7.1.1', 'datasette', 'requests', 'scipy', 'scikit-learn', 'matplotlib', 'nltk', 'textacy', 'gensim<=3.8.3', 'pyLDAvis' ],
    entry_points={ 'console_scripts': [ 'rdr = rdr.rdr:rdr' ] }
)



# datasette - given a carrel, browse its database

# require
from rdr import *

@click.command( options_metavar='[<options>]' )
@click.argument( 'carrel', metavar='<carrel>' )
def sql( carrel ) :

	"""Use Datasette to query <carrel>'s database
	
	Datasette is an browser-based interface to SQLite database files. Each study carrel includes such a file, and it contains information regarding parts-of-speech, named-entities, bibliographics, keywords, etc. Through the use of Datasette and a knowledge of SQL, the underlying database can be queried and results returned in ways not possible by merely browsing the given reports.
	
	Example: rdr datasette homer"""

	# configure
	DATASETTE = 'datasette'
		
	# require
	from os import system
	
	# sanity check
	checkForCarrel( carrel )

	# initialize
	localLibrary = configuration( 'localLibrary' )

	# build the shell command and go
	database = localLibrary/carrel/ETC/DATABASE
	system( DATASETTE + ' ' + str( database ) + ' -o' )





# require
from rdr import *

# config
@click.command()
@click.option('-f', '--file', default='stopwords', type=click.Choice( [ 'stopwords', 'foobar' ] ), help='edit the given file')
@click.argument( 'carrel', metavar='<carrel>' )
def edit( carrel, file ) :

	'''Edit the stopword list of the given <carrel>'''
	
	# sanity check
	checkForCarrel( carrel )

	# branch accordingly
	if file == 'stopwords' :
		file = str( configuration( 'localLibrary' )/carrel/ETC/STOPWORDS )
		click.edit( filename=file )